/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "sites.h"

extern struct site_t *all_sites;

extern char *copypath;

/* for the protocol driver structures */
extern struct proto_driver ftp_driver;

/* These are used for reporting errors back to the user. */
int rcfile_linenum; 
char *rcfile_line;

#ifdef USE_DAV
extern struct proto_driver dav_driver;
#endif

/* site_readsites will read ~/.sitecopyrc and fill the sites list.
 * This returns 0 on success, SITE_RCOPEN if the rcfile could not
 * be read, or SITE_RCCORRUPT if the rcfile was corrupt.
 * If it is corrupt, rcfile_linenum and rcfile_line are set to the
 * the corrupt line.
 */
int site_readsites( char *filename ) {
    FILE *fp;
    int state, last_state=8, ret=0;
    bool alpha, hash;
    char buf[BUFSIZ];
    char *ch, *ptr, key[BUFSIZ], val[BUFSIZ], val2[BUFSIZ];
    /* Holders for the site info */
    struct site_t *this_site, *last_site;
    struct name_mapping_t *new;

    if( (fp = fopen( filename, "r" )) == NULL ) {
	perror("fopen");
	return SITE_RCOPEN;
    } 

    last_site = this_site = NULL;
    rcfile_linenum = 0;
    rcfile_line = buf;

    while( fgets( buf, sizeof(buf), fp ) != NULL ) {
	rcfile_linenum ++;
	state = 0;
	ptr = key;
	memset( key, 0, BUFSIZ );
	memset( val, 0, BUFSIZ );
	memset( val2, 0, BUFSIZ );
	for( ch=buf; *ch!='\0'; ch++ ) {
	    alpha = !isspace(*ch); /* well, alphaish */
	    hash = (*ch == '#');
	    switch( state ) {
	    case 0: /* whitespace at beginning of line */
		if(hash) {
		    state = 8;
		} else if(alpha) {
		    *(ptr++) = *ch;
		    state = 1;
		}
		break;
	    case 1: /* key */
		if(hash) {
		    state = 8;
		} else if(!alpha) {
		    ptr = val;
		    state = 2;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 2: /* whitespace after key */
		if(hash) {
		    state = 8;
		} else if( *ch == '"' ) {
		    state = 4; /* begin quoted value */
		} else if ( alpha ) {
		    *(ptr++) = *ch;
		    state = 3;
		} 
		break;
	    case 3: /* unquoted value 1 */
		if(hash) {
		    state = 8;
		} else if( !alpha ) {
		    ptr = val2;
		    state = 5;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 4: /* quoted value 1 */
		if( *ch == '"' ) {
		    ptr = val2;
		    state = 5;
		} else if( *ch == '\\' ) {
		    last_state = 4;
		    state = 9;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 5: /* whitespace after value 1 */
		if(hash) {
		    state = 8;
		} else if( *ch == '"' ) {
		    state = 6; /* begin quoted value 2 */
		} else if ( alpha ) {
		    *(ptr++) = *ch;
		    state = 7; /* begin unquoted value 2 */
		} 
		break;
	    case 6: /* quoted value 2 */
		if( *ch == '"' ) {
		    state = 8;
		} else if( *ch == '\\' ) {
		    last_state = 4;
		    state = 9;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 7: /* unquoted value 2 */
		if(hash) {
		    state = 8;
		} else if( !alpha ) {
		    state = 8;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 8: /* ignore till end of line */
		break;
	    case 9: /* a literal (\-slashed) in a value */
		*(ptr++) = *ch;
		state = last_state;
		break;
	    }
	}
	
	DEBUG( DEBUG_RCFILE, "Key [%s] Value: [%s] Value2: [%s]\n", key, val, val2);
	
	if( strlen( key ) == 0 ) {
	    /* Blank line */
	    continue;
	}
	if( strlen( val ) == 0 ) {
	    /* A key with no value. */
	    if( this_site == NULL ) {
		/* Need to be in a site! */
		ret = SITE_RCCORRUPT;
		break;
	    }
	    if( strcmp( key, "nodelete" ) == 0 ) {
		this_site->nodelete = true;
	    } else if( strcmp( key, "checkmoved" ) == 0 ) {
		this_site->checkmoved = true;
	    } else {
		ret = SITE_RCCORRUPT;
		break;
	    }
	} else if( strlen( val2 ) == 0 ) {
	    /* A key with a single value. */
	    if( strcmp( key, "site" ) == 0 ) {
		/* Beginning of a new Site */
		last_site = this_site;
		/* Allocate new item */
		this_site = malloc( sizeof( struct site_t ) );
		/* Zero it out */
		memset( this_site, 0, sizeof( struct site_t ) );
		if( last_site != NULL ) { /* next site */
		    last_site->next = this_site;
		} else { /* First site */
		    all_sites = this_site;
		}
		this_site->name = strdup( val );
		this_site->files = NULL;
		/* defaults... */
		this_site->perms = sitep_ignore;
		this_site->symlinks = sitesym_follow;
		this_site->nodelete = false;
		this_site->checkmoved = false;
		this_site->driver = &ftp_driver;
		this_site->ftp_pasv_mode = true;
		/* Now work out the info filename */
		this_site->infofile = malloc( strlen(copypath) + strlen(val) + 1);
		strcpy( this_site->infofile, copypath );
		strcat( this_site->infofile, val );
	    } else if( this_site == NULL ) {
		ret = SITE_RCCORRUPT;
		break; 
	    } else if( strcmp( key, "username" ) == 0 ) {
		/* username */
		this_site->username = strdup( val );
	    } else if( strcmp( key, "server" ) == 0 ) {
		this_site->server = strdup( val );
	    } else if( strcmp( key, "port" ) == 0 ) {
		this_site->port = atoi( val );
	    } else if( strcmp( key, "password" ) == 0 ) {
		this_site->password = strdup( val );
	    } else if( strcmp( key, "url" ) == 0 ) {
	        this_site->url = strdup( val );
	    } else if( strcmp( key, "remote" ) == 0 ) {
		/* Relative filenames must start with "~/" */
		if( val[0] == '~' ) {
		    if( strlen(val) > 1 ) {
			if( val[1] == '/' ) {
			    if( val[strlen(val)-1] != '/' )
				strcat( val, "/" );
			    this_site->remote_root = strdup( val + 2 );
			} else {
			    ret = SITE_RCCORRUPT;
			    break; 
			}
		    } else {
			ret = SITE_RCCORRUPT;
			break;
		    }
		} else { /* Dirname doesn't begin with a ~ */
		    if( val[strlen(val)-1] != '/' )
			strcat( val, "/" );
		    this_site->remote_root = strdup( val );
		}
	    } else if( strcmp( key, "local" ) == 0 ) {
		if( val[strlen(val)-1] != '/' )
		    strcat( val, "/" );
		this_site->local_root = strdup( val );
	    } else if( strcmp( key, "permissions") == 0 ) {
		if( strcmp( val, "ignore" ) == 0 ) {
		    this_site->perms = sitep_ignore;
		} else if( strcmp( val, "exec" ) == 0 ) {
		    this_site->perms = sitep_exec;
		} else if( strcmp( val, "all" ) == 0 ) {
		    this_site->perms = sitep_all;
		} else {
		    ret = SITE_RCCORRUPT;
		    break;
		}
	    } else if( strcmp( key, "symlinks" ) == 0 ) {
		if( strcmp( val, "follow" ) == 0 ) {
		    this_site->symlinks = sitesym_follow;
		} 
#if 0 /* not yet - can we do this with FTP? */
		else if( strcmp( val, "maintain" ) == 0 ) {
		    this_site->symlinks = sitesym_maintain;
		}
#endif
		else if( strcmp( val, "ignore" ) == 0 ) {
		    this_site->symlinks = sitesym_ignore;
		} else {
		    ret = SITE_RCCORRUPT;
		    break;
		}
	    } else if( strcmp( key, "exclude" ) == 0 ) {
		/* Find the position in the array */
		this_site->excludes[this_site->numexcludes] = strdup(val);
		this_site->numexcludes++;
	    } else if( strcmp( key, "protocol" ) == 0 ) {
		if( strcmp( val, "ftp" ) == 0 ) {
		    this_site->protocol = siteproto_ftp;
		    this_site->driver = &ftp_driver;
		}
#ifdef USE_DAV
		else if( strcmp( val, "http" ) == 0 ) {
		    this_site->protocol = siteproto_http;
		    this_site->driver = &dav_driver;
		}
#endif /* USE_DAV */
	        else {
		    ret = SITE_RCCORRUPT;
		    break;
		}
	    } else if( strcmp( key, "ftp" ) == 0 ) {
		if( strcmp( val, "nopasv" ) == 0 ) {
		    this_site->ftp_pasv_mode = false;
		} else {
		    ret = SITE_RCCORRUPT;
		}
	    } else {
		/* Unknown key! */
		ret = SITE_RCCORRUPT;
		break;
	    }
	} else {
	    /* Key with two values! */
	    if( strcmp( key, "rename" ) == 0 ) {
		new = malloc( sizeof(struct name_mapping_t) );
		new->next = this_site->renames;
		this_site->renames = new;
		new->local = strdup( val );
		new->remote = strdup( val2 );
	    } else {
		ret = SITE_RCCORRUPT;
		break; 
	    }
	}
    }

    fclose( fp );
    return ret;
}
