/* 
   socket handling routines
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>, except where
   otherwise indicated.
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SOCKET_H
#define SOCKET_H

#include <sys/socket.h>
#include <sys/types.h>

#include <netinet/in.h>

/* Dump the given filename down the given socket
 * Returns non-zero value if successful */
int send_file( const int sock, const char *filename ); 
/* As above, backwards */
int recv_file( const int sock, const char *filename, const size_t filesize ); 

/* Chief bulk data transfer bod */
size_t transfer( const int fromfd, const int tofd, size_t readlen );

/* Sends the given line to given socket, CRLF appended */
int send_line( const int sock, const char *line ); 
/* Sends the given block of data down the socket */
int send_data( const int sock, const char *data, const size_t length ); 
/* Sends the null-terminated string down the given socket */
int send_string( const int sock, const char *string ); 

/* Reads a line from given socket */
int read_line( int sock, char *line, int len ); 
int read_data( const int sock, char *buffer, int buflen );

/* Creates and connects a socket */
int socket_connect( const struct in_addr host, const int portnum );

void socket_close( const int socket );

int host_lookup( const char *hostname, struct in_addr *addr );

/* Returns the standard TCP port for the given service */
int get_tcp_port( const char *name );

#endif /* SOCKET_H */
